# -*- coding: utf-8 -*-
"""
Created on Mon Oct  7 10:03:02 2019

@author: icruicks
"""

import numpy as np
from sklearn.preprocessing import normalize
from sknetwork.clustering import Louvain, modularity
from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics import silhouette_score as sils
from random import shuffle
from scipy.sparse import csr_matrix, identity, csc_matrix
from scipy.sparse.linalg import inv
from scipy.spatial.distance import cdist


class CMIC:
    
    def __init__(self, alpha=0.9, iterations=10, final_cluster_iterations=5,
                 iter_thresh = 0, epsilon=0.5):
        
        self.iterations = iterations
        self.alpha = alpha
        self.W = []
        self.L = []
        self.A = []
        self.final_cluster_iterations = final_cluster_iterations
        self.iter_thresh = iter_thresh
        self.epsilon = epsilon
        self.inverses = []
    
    def fit_predict(self, base_clusters, modal_graphs):
        ba_matrices =[]
        idxs = []
        idxs_iter = 0
        
        for base_cluster in base_clusters:
            ba_matrix = np.zeros((base_cluster.size, base_cluster.max()+1))
            ba_matrix[np.arange(base_cluster.size), base_cluster] = 1
            ba_matrices.append(ba_matrix)
            idxs.append([idxs_iter, idxs_iter+ba_matrix.shape[1]])
            idxs_iter += ba_matrix.shape[1]
            
        ba_matrix = np.concatenate(ba_matrices, axis=1)
        
        
        for G in modal_graphs:
            D = np.diag(np.ravel(np.sum(G, axis=1)))
            self.L.append(D-G)
            self.A.append(G)
            self.W.append(normalize(G.T, norm='l1', axis=1, copy=True))
            # self.inverses.append(csr_matrix(inv(csc_matrix(identity(G.shape[0])-self.alpha * normalize(G.T, norm='l1', axis=1, copy=True)))))
             
            
        #self.alpha = [0.99, 0.55]
            
        ca_matrix = ba_matrix.copy()
        # ca_matrix_prev = ba_matrix.copy()

        for t in range(self.iterations):
            #shuffle(self.W)
            # curr_epsilon = 1-self.epsilon ** t
            temp = []
            # O = 1-cdist(ca_matrix, ca_matrix, metric='cosine')
            # O[O < curr_epsilon] = 0
            # O[O > 0] = 1
            
            ''' for individualized weightings '''
            # for m in range(len(self.W)):
            #     G_m = self.W[m]
            #     I = np.eye(G_m.shape[0])
            #     a = np.max(np.abs(self.L[m] @ ca_matrix).T/np.diff(G_m.indptr), axis=0)
            #     A = np.diag(np.ravel(a))
            #     temp.append(A @ G_m @ ca_matrix + (I-A) @ ba_matrix)
            
            ''' for use with similarity bias '''
            # for m in range(len(self.W)):
            #     O = 1-cdist(ca_matrix[:,idxs[m][0]:idxs[m][1]], ca_matrix[:,idxs[m][0]:idxs[m][1]], metric='cosine')
            #     O[O < self.epsilon] = 0
            #     O[O > 0] = 1
            #     # A_m_t = self.A[m].multiply(csr_matrix(O))
            #     # G_m = normalize(A_m_t.T, norm='l1', axis=1, copy=True)
            #     G_m = normalize(O.T, norm='l1', axis=1, copy=True)
            #     # G_m = self.W[m].multiply(csr_matrix(O))
            #     # temp.append(self.alpha * G_m @ ca_matrix + (1-self.alpha)* ca_matrix_prev)
            #     temp.append(self.alpha * G_m @ ca_matrix + (1-self.alpha)* ca_matrix)
            #     # temp.append(self.alpha * G_m @ ca_matrix + (1-self.alpha)* ba_matrix)
            
            ''' for use with implicit scheme '''
            # for m in range(len(self.W)):
                # O = 1-cdist(ca_matrix[:,idxs[m][0]:idxs[m][1]], ca_matrix[:,idxs[m][0]:idxs[m][1]], metric='cosine')
                # O[O < self.epsilon[m]] = 0
                # O[O > 0] = 1
                # A_m_t = self.A[m].multiply(csr_matrix(O))
                # G_m = normalize(A_m_t.T, norm='l1', axis=1, copy=True)
                # inverse = csr_matrix(inv(csc_matrix(identity(G_m.shape[0])-self.alpha * G_m)))
                # temp.append(inverse*((1-self.alpha)*ca_matrix))
                # # temp.append(self.inverses[m]*((1-self.alpha)*ca_matrix))
            
            
            ''' for use when you only want influence to apply to off current mode C's '''   
#            for m in range(len(self.W)):
#                G_m = self.W[m]
#                c = ca_matrix.copy()
#                c[:,idxs[m][0]:idxs[m][1]+1] =0 
#                c_0 =ba_matrix.copy()
#                c_0[:,idxs[m][0]:idxs[m][1]+1] = 0
#                modal_ca_matrix = self.alpha * G_m @ c + (1-self.alpha)* c_0
#                temp.append(modal_ca_matrix)
             # ca_matrix = np.sum(temp, axis=0)/ (len(self.W)-1) 
            
            ''' for use in the base model '''
            for G_m in self.W:
                # modal_ca_matrix = self.alpha * G_m @ ca_matrix + (1-self.alpha)* ba_matrix 
                # modal_ca_matrix = np.concatenate([self._get_max_entry(modal_ca_matrix[:,mode_idx[0]:mode_idx[1]+1]) for mode_idx in idxs], axis=1)
                # temp.append(modal_ca_matrix)
                temp.append(self.alpha * G_m @ ca_matrix + (1-self.alpha)* ba_matrix )
                # temp.append(self.alpha * G_m @ ca_matrix + (1-self.alpha)* ca_matrix_prev)
            
            ''' for doing more influence iterations before a cross iteration '''
            # for sub_iteration in range(50):
            #     for m in range(len(temp)):
            #         G_m = self.W[m]
            #         temp[m] = self.alpha * G_m @ temp[m] + (1-self.alpha)* ba_matrix 

            ''' for use with different modal alphas '''
#            for idx in range(len(self.W)):
#                G_m = self.W[idx]
#                alpha = self.alpha[idx]
#                temp.append(alpha * G_m @ ca_matrix + (1-alpha)* ba_matrix )
            
            ''' for use with a weighting in the cross scheme '''
            # beta =[]
            # for m in range(len(self.W)):
            #     G_m = self.W[m]
            #     C_m = self.alpha * G_m @ ca_matrix + (1-self.alpha)* ba_matrix 
            #     temp.append(C_m)
            #     beta_m = np.mean([np.trace(C_m.T @ self.L[mode] @ C_m) for mode in range(len(self.W))])
            #     beta.append(beta_m)
            # beta = np.min(beta) / np.array(beta)
            # ca_matrix = np.sum([beta[m] * temp[m] for m in range(len(self.W))], axis=0) / np.sum(beta)
            
            # ca_matrix_prev = ca_matrix.copy()
                
            ca_matrix = np.mean(temp, axis=0)
            
            # clusters = np.concatenate(temp, axis=1)
            # new_clusters = Louvain().fit(clusters).labels_
            # ca_matrix = np.zeros((new_clusters.size, new_clusters.max()+1))
            # ca_matrix[np.arange(new_clusters.size), new_clusters] = 1
#            self.alpha *=0.9
            #edges = ca_matrix[np.nonzero(ca_matrix)]
            #threshold = np.quantile(edges, self.iter_thresh)
            #ca_matrix[ca_matrix < threshold] = 0
#            ca_matrix[ca_matrix < 0.1] = 0
            #ca_matrix = np.concatenate([normalize(ca_matrix[:,mode_idx[0]:mode_idx[1]+1], norm='l1', axis=1, copy=True) for mode_idx in idxs], axis=1)
#            ca_matrix = np.concatenate([self._get_max_entry(ca_matrix[:,mode_idx[0]:mode_idx[1]+1]) for mode_idx in idxs], axis=1)
        
        return ca_matrix

    
    def _get_max_entry(self, a):
        b = np.zeros_like(a)
        b[np.arange(len(a)), a.argmax(1)] = a.max(axis=1)
        return b
    
    def _DSM(self, graph):
        graph = graph.todense()
        graph = normalize(graph, norm='l1', axis=1, copy=False)
        col_sums= np.sum(graph, axis=0)
        temp_graph = np.divide(graph, col_sums**0.5, out=np.zeros_like(graph), where=col_sums!=0)
        return csr_matrix(temp_graph @ temp_graph.T)

        
class CMIWCC:
    
    def __init__(self, alpha=0.9, iterations=10):
        
        self.iterations = iterations
        self.alpha = alpha
        self.W = []
    
    def fit_predict(self, base_clusters, modal_graphs):
        ba_matrices =[]
        idxs = []
        idxs_iter = 0
        
        for base_cluster in base_clusters:
            ba_matrix = np.zeros((base_cluster.size, base_cluster.max()+1))
            ba_matrix[np.arange(base_cluster.size), base_cluster] = 1
            ba_matrices.append(ba_matrix)
            idxs.append([idxs_iter, idxs_iter+ba_matrix.shape[1]])
            idxs_iter += ba_matrix.shape[1]
            
        ba_matrix = np.concatenate(ba_matrices, axis=1)
        
        for G in modal_graphs:
            self.W.append(normalize(G, norm='l1', axis=1, copy=True))
            
        C = np.zeros((ba_matrix.shape[1], ba_matrix.shape[1]))
        for col_idx in range(ba_matrix.shape[1]):
            for idx in idxs:
                if idx[0] <= col_idx < idx[1]:
                    pass
                else:
                    for alt_idx in range(idx[0], idx[1]):
                        if C[col_idx, alt_idx] <= 0:
                            C[col_idx, alt_idx] = np.dot(ba_matrix[:,col_idx], ba_matrix[:,alt_idx])/ np.count_nonzero(ba_matrix[:,alt_idx])

        # H = C.copy()
        # H[np.nonzero(H)] = -H[np.nonzero(H)]*np.log2(H[np.nonzero(H)])
        # C_diags = np.diag(np.exp(-np.sum(H,axis=1)/(0.5*len(modal_graphs))))
        # C = normalize(C + C_diags, norm='l1', axis=1, copy=True)
        C = normalize(C, norm='l1', axis=1, copy=True)
        
        ca_matrix = ba_matrix.copy()
        for t in range(self.iterations):
            temp = []

            for G_m in self.W:
                temp.append(self.alpha * G_m @ ca_matrix @ C.T + (1-self.alpha)* ba_matrix )
                
            ca_matrix = np.mean(temp, axis=0)

        return ca_matrix
        
        
    
class MSIMC:
    
    def __init__(self, lambda_power =1.2, ro=0.4, iterations=4, n_clusters=None):
        
        self.lmda = lambda_power
        self.ro = ro
        self.T = iterations
        self.n_clusters = n_clusters
        
    def fit_predict(self, modal_graphs):
        self.R = identity(modal_graphs[0].shape[0])
        self.C = [identity(modal_graphs[i].shape[1]) for i in range(len(modal_graphs))]
        self.M = [graph.power(self.lmda) for graph in modal_graphs]
        
        for iteration in range(self.T):
            C_t_m = [self.M[i].T @ self.R @ self.M[i] for i in range(len(self.M))]
            C_t_m = [self._pseudo_scale(C) for C in C_t_m]
            C_t_m = [self._prune(C) for C in C_t_m]
            R_t_m = [self.M[i] @ self.C[i] @ self.M[i].T for i in range(len(self.M))]
            R_t_m = [self._pseudo_scale(R) for R in R_t_m]
            R_t_m = [self._prune(R) for R in R_t_m]
            clusters = [Louvain().fit(R).labels_ 
                        for R in R_t_m]
            modularities = [modularity(R, cluster) for R, cluster in zip(R_t_m, clusters)]
            self.R =self._create_new_R(R_t_m[np.argmax(modularities)], clusters)
            self.C = C_t_m
            # clusters = [AgglomerativeClustering(n_clusters = self.n_clusters, affinity='precomputed', linkage='average').fit(1-R.todense()).labels_ 
            #             for R in R_t_m]
            # R_t_m = [R.todense() for R in R_t_m]
            # [np.fill_diagonal(R, 1) for R in R_t_m]
            # goodness = [sils(1-R, cluster, metric='precomputed') for R, cluster in zip(R_t_m, clusters)]
            # self.R =self._create_new_R(csr_matrix(R_t_m[np.argmax(goodness)]), clusters)

        return Louvain().fit(self.R).labels_
        # return AgglomerativeClustering(n_clusters = self.n_clusters, affinity='precomputed', linkage='average').fit(1-self.R.todense()).labels_
        
    def _pseudo_scale(self, matrix):
        result = matrix.power(1/self.lmda).todense()
        normalization_term = np.power(np.outer(matrix.diagonal(), matrix.diagonal()), 1/(2*self.lmda))
        return csr_matrix(np.clip(np.divide(result, normalization_term, out=np.zeros_like(result), where=normalization_term!=0),
                                  0,1)) 
    
    def _prune(self, matrix):
        threshold = np.quantile(matrix.data, self.ro)
        matrix.data[matrix.data < threshold] = 0
        matrix.eliminate_zeros()
        return matrix
    
    
    def _create_new_R(self, matrix, clusters):
        B = []
        for cluster in clusters:
            ba_matrix = np.zeros((cluster.size, cluster.max()+1))
            ba_matrix[np.arange(cluster.size), cluster] = 1
            B.append(ba_matrix)
        B = np.concatenate(B, axis=1)
        R_new = B @ B.T
        mask_1 = R_new != 0
        R_new[R_new < len(clusters)] =0
        R_new[R_new > 0] =1
        mask_2 = R_new !=1
        return csr_matrix(R_new + np.multiply(mask_2, np.multiply(mask_1, matrix.todense())))
        
           
    
class CND:
    def __init__(self, iterations=10, alpha=0.9, iter_thresh=0, interior_term ='clusters',
                 secondary_term='identity'):
        
        self.iter_thresh = iter_thresh
        self.T = iterations
        self.alpha = alpha
        self.interior_term = interior_term
        self.secondary_term = secondary_term
        
        
    def fit_predict(self, clusters, modal_graphs):
        B = []
        CA = []
        for cluster in clusters:
            ba_matrix = np.zeros((cluster.size, cluster.max()+1))
            ba_matrix[np.arange(cluster.size), cluster] = 1
            B.append(ba_matrix)
            CA.append(csr_matrix(normalize(ba_matrix @ ba_matrix.T, norm='l1', axis=1, copy=True)))
        B = np.concatenate(B, axis=1)         
        
        self.kernel_graphs = [normalize(graph, norm='l1', axis=1, copy=True) for graph in modal_graphs]
        
        if self.interior_term == 'identity':    
            # self.P = identity(modal_graphs[0].shape[0]) 
            self.P = [identity(modal_graphs[0].shape[0]) for _ in range(len(modal_graphs))]
        else:
            self.P = CA.copy()
        
        # CA = csr_matrix(normalize(1/(len(clusters)) * B@ B.T, norm='l1', axis=1, copy=True))
        # CA = csr_matrix(B@B.T)
        # CA.data[CA.data < len(clusters)] =0
        # CA.eliminate_zeros()
        # CA = normalize(CA, norm='l1', axis=1, copy=True)
        
        # self.P = CA.copy()
        # self.P = B.copy()
        # self.C = [identity(modal_graphs[0].shape[1]) for _ in range(len(modal_graphs))]
        # self.C = identity(modal_graphs[0].shape[1])
        
        for iteration in range(self.T):
            P_new = []
            # C_new =[]
            for v in range(len(self.kernel_graphs)):
                S_v = self.kernel_graphs[v]
                
                P_k = np.sum([self.P[idx] for idx in range(len(self.P)) if idx != v], axis=0)/(len(self.P)-1)
                if self.secondary_term == 'identity':
                    result = self.alpha * (S_v @ P_k @ S_v.T) + (1-self.alpha) * identity(modal_graphs[0].shape[0])
                else:
                    result = self.alpha * (S_v @ P_k @ S_v.T) + (1-self.alpha) * CA[v]
                P_new.append(self._prune(result))
                
                # result = self.alpha * (S_v @ self.P @ S_v.T) + (1-self.alpha) * identity(modal_graphs[0].shape[0])
                # result = self.alpha * (S_v @ self.P @ S_v.T) + (1-self.alpha) * CA
                # P_new.append(self._prune(result))
                
                # C_result = self.alpha * (S_v.T @ self.P @ S_v) + (1-self.alpha) * identity(modal_graphs[0].shape[0])
                # P_result = self.alpha * (S_v @ self.C[v] @ S_v.T) + (1-self.alpha) * identity(modal_graphs[0].shape[0])
                # P_result = self.alpha * (S_v @ self.C @ S_v.T) + (1-self.alpha) * identity(modal_graphs[0].shape[0])
                # P_new.append(self._prune(P_result))
                # C_new.append(csr_matrix(self._prune(C_result)))
                
                # C_k = np.sum([self.C[idx] for idx in range(len(self.C)) if idx != v], axis=0)/(len(self.C)-1)
                # P_k = np.sum([self.P[idx] for idx in range(len(self.P)) if idx != v], axis=0)/(len(self.P)-1)
                # P_result = self.alpha * (S_v @ C_k @ S_v.T) + (1-self.alpha) * identity(modal_graphs[0].shape[0])
                # C_result = self.alpha * (S_v.T @ P_k @ S_v) + (1-self.alpha) * identity(modal_graphs[0].shape[0])
                # P_new.append(self._prune(P_result))
                # C_new.append(csr_matrix(self._prune(C_result)))
                
            # clusters = [Louvain().fit(P).labels_ for P in P_new]
            # modularities = [modularity(P, cluster) for P, cluster in zip(P_new, clusters)]
            # self.P =self._create_new_P(P_new[np.argmax(modularities)], clusters)
            # self.P = np.sum(P_new, axis=0)/(len(modal_graphs))
            # self.C = np.sum(P_new, axis=0)/(len(modal_graphs))
            # clusters_C = [Louvain().fit(C).labels_ for C in C_new]
            # modularities_C = [modularity(C, cluster) for C, cluster in zip(C_new, clusters_C)]
            # self.C =self._create_new_P(C_new[np.argmax(modularities_C)], clusters_C)
            # self.C = C_new
            self.P = P_new
                
        self.P = np.sum(P_new, axis=0)/(len(modal_graphs))
        return Louvain().fit(self.P).labels_
        
    
    def _prune(self, matrix):
        threshold = np.quantile(matrix.data, self.iter_thresh)
        matrix.data[matrix.data < threshold] = 0
        matrix.eliminate_zeros()
        return matrix
        
        
    def _create_new_P(self, matrix, clusters):
        B = []
        for cluster in clusters:
            ba_matrix = np.zeros((cluster.size, cluster.max()+1))
            ba_matrix[np.arange(cluster.size), cluster] = 1
            B.append(ba_matrix)
        B = np.concatenate(B, axis=1)
        R_new = B @ B.T
        mask_1 = R_new != 0
        R_new[R_new < len(clusters)] =0
        R_new[R_new > 0] =1
        mask_2 = R_new !=1
        return csr_matrix(normalize(R_new + np.multiply(mask_2, np.multiply(mask_1, matrix.todense())), norm='l1', axis=1, copy=True))
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        